/* sstrip, version 1.0: Copyright (C) 1999 by Brian Raiter, under the
 * GNU General Public License. No warranty. See COPYING for details.
 */

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<stdarg.h>
#include	<errno.h>
#include	<unistd.h>
#include	<linux/elf.h>

#ifndef TRUE
#define	TRUE		1
#define	FALSE		0
#endif

/* The memory-allocation macro.
 */
#define	alloc(p, n)	(((p) = realloc(p, n))				\
			    || (fputs("Out of memory.\n", stderr),	\
				exit(EXIT_FAILURE), 0))

/* This structure tracks pieces of the original file that need to be
 * copied into the new file.
 */
typedef	struct pieceinfo {
    Elf32_Word	from;		/* start of a piece to save		*/
    Elf32_Word	to;		/* end of a piece to save		*/
    Elf32_Word	size;		/* (to - from)				*/
    Elf32_Word	offset;		/* where it will go in the new file	*/
} pieceinfo;

static char const      *thefilename;	/* the current file name	*/
static FILE	       *thefile;	/* the current file handle	*/

static Elf32_Ehdr	elfhdr;		/* original ELF header		*/
static Elf32_Phdr      *phdrs = NULL;	/* original program header tbl	*/
static unsigned long	phdrsize;	/* size of program header tbl	*/

static pieceinfo       *pieces = NULL;	/* all the pieces to copy over	*/
static int		piecenum;	/* the number of pieces		*/

static char	       *newbuf = NULL;	/* buffer for the new file	*/
static unsigned long	newsize;	/* size of the new file		*/

/* An error-handling function. The given error message is used only
 * when errno is not set.
 */
static int err(char const *errmsg)
{
    if (errno)
	perror(thefilename);
    else
	fprintf(stderr, "%s: %s\n", thefilename, errmsg);
    return FALSE;
}

/* readheaders() reads the ELF header and, after various checks to
 * make sure that this is in fact a file that we should be munging,
 * the program header table.
 */
static int readheaders(void)
{
    int		bigend;

    errno = 0;
    if (fread(&elfhdr, sizeof elfhdr, 1, thefile) != 1)
	return err("not an ELF file.");
    if (elfhdr.e_ident[EI_MAG0] != ELFMAG0
		|| elfhdr.e_ident[EI_MAG1] != ELFMAG1
		|| elfhdr.e_ident[EI_MAG2] != ELFMAG2
		|| elfhdr.e_ident[EI_MAG3] != ELFMAG3)
	return err("not an ELF file.");

    bigend = TRUE;
    *(char*)&bigend = 0;
    if (elfhdr.e_ident[EI_DATA] != (bigend ? ELFDATA2MSB : ELFDATA2LSB)) {
	fprintf(stderr, "%s: not %s-endian.\n",
			thefilename, bigend ? "big" : "little");
	return FALSE;
    }
    if (elfhdr.e_ehsize != sizeof(Elf32_Ehdr)) {
	fprintf(stderr, "%s: unrecognized ELF header size "
			"(size = %u instead of %u).\n",
			thefilename, elfhdr.e_ehsize, sizeof(Elf32_Ehdr));
	return FALSE;
    }
    if (!elfhdr.e_phoff)
	return err("no program header table.");
    if (elfhdr.e_phentsize != sizeof(Elf32_Phdr)) {
	fprintf(stderr, "%s: unrecognized program header size "
			"(size = %u instead of %u).\n",
			thefilename, elfhdr.e_phentsize, sizeof(Elf32_Ehdr));
	return FALSE;
    }
    if (elfhdr.e_type != ET_EXEC && elfhdr.e_type != ET_DYN)
	return err("not a program file.");

    phdrsize = elfhdr.e_phnum * elfhdr.e_phentsize;
    alloc(phdrs, phdrsize);
    errno = 0;
    if (fread(phdrs, phdrsize, 1, thefile) != 1)
	return err("invalid program header table.");

    return TRUE;
}

/* truncatezeros() examines the ends of the contents of each loadable
 * segment, and reduces their file size to leave out trailing NULs.
 */
static int truncatezeros(void)
{
    static char	buf[1024];
    Elf32_Word	count;
    int		i, n;

    for (i = 0 ; i < elfhdr.e_phnum ; ++i) {
	if (phdrs[i].p_type != PT_LOAD && phdrs[i].p_type != PT_DYNAMIC)
	    continue;
	count = phdrs[i].p_filesz;
	n = -1;
	while (count > 0 && n < 0) {
	    n = count > sizeof buf ? sizeof buf : count;
	    if (fseek(thefile, phdrs[i].p_offset + count - n, SEEK_SET)
			|| fread(buf, n, 1, thefile) != 1)
		return err("invalid program header table");
	    while (n-- && !buf[n])
		--count;
	}
	if (phdrs[i].p_filesz - count > 3)
	    phdrs[i].p_filesz = count;
    }

    return TRUE;
}

/* A function for use with qsort.
 */
static int piecesorter(const void *s1, const void *s2)
{
    return ((pieceinfo*)s1)->from - ((pieceinfo*)s2)->from;
}

/* findpieces() locates all the parts of the file that are pointed to
 * by program header entries (excepting any entries of type PT_NOTE,
 * which are discarded), and assembles them into a sorted list of
 * non-overlapping regions.
 */
static int findpieces(void)
{
    pieceinfo  *p1, *p2;
    int		same;
    int		i, j;

    piecenum = elfhdr.e_phnum + 2;
    alloc(pieces, piecenum * sizeof *pieces);
    for (i = 0, p1 = pieces ; i < elfhdr.e_phnum ; ++i, ++p1) {
	if (phdrs[i].p_type == PT_NOTE) {
	    memset(phdrs + i, 0, sizeof *phdrs);
	    phdrs[i].p_type = PT_NULL;
	}
	if (phdrs[i].p_type == PT_NULL)
	    p1->from = p1->to = 0;
	else {
	    p1->from = phdrs[i].p_offset;
	    p1->to = phdrs[i].p_offset + phdrs[i].p_filesz;
	}
    }
    p1->from = 0;
    p1->to = sizeof(Elf32_Ehdr);
    ++p1;
    p1->from = elfhdr.e_phoff;
    p1->to = elfhdr.e_phoff + phdrsize;

    qsort(pieces, piecenum, sizeof *pieces, piecesorter);

    do {
	same = TRUE;
	for (i = 0, p1 = pieces ; same && i < piecenum - 1 ; ++i, ++p1) {
	    for (j = i + 1, p2 = p1 + 1 ; same && j < piecenum ; ++j, ++p2) {
		if (p1->to >= p2->from) {
		    if (p1->to < p2->to)
			p1->to = p2->to;
		    --piecenum;
		    memmove(p2, p2 + 1, (piecenum - j) * sizeof *pieces);
		    same = FALSE;
		}
	    }
	}
    } while (!same);

    for (i = 0, p1 = pieces ; i < piecenum ; ++i, ++p1)
	p1->size = p1->to - p1->from;
    pieces[0].offset = 0;
    for (i = 1, p1 = pieces ; i < piecenum ; ++i, ++p1) {
	p1->size = p1->to - p1->from;
	p1[1].offset = (p1->offset + p1->size + 3) & ~3UL;
    }
    p1->size = p1->to - p1->from;

    return TRUE;
}

/* modifyheaders() removes all references to the section header table,
 * and then shifts the offsets in the program header table in
 * accordance with the organization of the stripped file-to-be.
 */
static int modifyheaders(void)
{
    int	i, j;

    elfhdr.e_shoff = 0;
    elfhdr.e_shnum = 0;
    elfhdr.e_shentsize = 0;
    elfhdr.e_shstrndx = 0;

    for (i = 0 ; i < elfhdr.e_phnum ; ++i) {
	if (phdrs[i].p_type == PT_NULL)
	    continue;
	for (j = 0 ; j < piecenum ; ++j)
	    if (phdrs[i].p_offset >= pieces[j].from
				&& phdrs[i].p_offset < pieces[j].to)
		break;
	if (j == piecenum) {
	    if (phdrs[i].p_filesz)
		return err("invalid program header table");
	    j = piecenum - 1;
	    phdrs[i].p_offset = pieces[j].to;
	}
	phdrs[i].p_offset = pieces[j].offset
			    + (phdrs[i].p_offset - pieces[j].from);
    }

    for (j = 0 ; j < piecenum ; ++j)
	if (elfhdr.e_phoff >= pieces[j].from && elfhdr.e_phoff < pieces[j].to)
	    break;
    elfhdr.e_phoff = pieces[j].offset + (elfhdr.e_phoff - pieces[j].from);

    return TRUE;
}

/* buildstripped() loads the pieces of the executable into a single
 * buffer, and replaces the original ELF header and program header
 * table with the new version.
 */
static int buildstripped(void)
{
    unsigned long	n;
    int			i;

    newsize = pieces[piecenum - 1].offset + pieces[piecenum - 1].size;
    alloc(newbuf, newsize);

    for (i = n = 0 ; i < piecenum ; ++i) {
	if (!pieces[i].size)
	    continue;
	if (pieces[i].offset > n) {
	    memset(newbuf + n, 0, pieces[i].offset - n);
	    n = pieces[i].offset;
	}
	errno = 0;
	if (fseek(thefile, pieces[i].from, SEEK_SET)
			|| fread(newbuf + n, pieces[i].size, 1, thefile) != 1)
	    return err("invalid program header table");
	n += pieces[i].size;
    }

    *(Elf32_Ehdr*)newbuf = elfhdr;
    memcpy(newbuf + elfhdr.e_phoff, phdrs, phdrsize);

    return TRUE;
}

/* savestripped() writes the new contents back to the original file
 * and sets the file length.
 */
static int savestripped(void)
{
    rewind(thefile);

    errno = 0;
    if (fwrite(newbuf, newsize, 1, thefile) != 1
		|| ftruncate(fileno(thefile), newsize)) {
	err("could not write to contents");
	fprintf(stderr, "WARNING: %s may be corrupted!\n", thefilename);
	return FALSE;
    }

    return TRUE;
}

/* main() loops over the cmdline arguments, leaving all the real work
 * to the other functions.
 */
int main(int argc, char *argv[])
{
    char	      **arg;
    int			ret = 0;

    if (argc < 2 || !strcmp(argv[1], "-h")) {
	printf("sstrip, version 1.0: Copyright (C) 1999 Brian Raiter\n"
	       "Usage: sstrip FILE...\n");
	return 0;
    }

    for (arg = argv + 1 ; (thefilename = *arg) != NULL ; ++arg) {
	if (!(thefile = fopen(thefilename, "rb+"))) {
	    err("unable to open.");
	    ++ret;
	    continue;
	}
	if (!readheaders() || !truncatezeros() || !findpieces()
			   || !modifyheaders() || !buildstripped()
			   || !savestripped())
	    ++ret;
	fclose(thefile);
    }

    return ret;
}
